/*  *********************************************************************** */
/*  *********************************************************************** */

/*  FILE edp.c - *SAT 1.0 */
/*  Consistency check in logic E */

/*  *********************************************************************** */
/*  *********************************************************************** */

#include "edp.h"

/* *********************************************************************** */
/* The function for testing E consistency of an assignment */
/* *********************************************************************** */
int Test_E_consistency(image_t *cur_image_ptr, dag_t *dag_ptr)
{
  int          result, test_ab, test_ba;
  lsStatus     list_status;
  st_table     *agents_st_table_ptr;
  lsList       *alpha_lsList_ptr, *beta_lsList_ptr;
  lsList       *to_test_lsList_ptr;
  st_generator *at_st_generator;
  lsGen        alpha_lsGen, beta_lsGen;
  stGeneric    ak_stGeneric, av_stGeneric;
  lsGeneric    alpha_lsGeneric;
  lsGeneric    beta_lsGeneric;
  vertex_t     *alpha_vertex_ptr, *beta_vertex_ptr;
  image_t      *new_image_ptr;

#ifdef TRACE_CONS
  printf("------------------------------");
  printf("Considering the model:\n");
  print_model_complete(cur_image_ptr);
  printf("\n");
#endif
  /* This is the classification for agents */
  agents_st_table_ptr = st_init_table(st_numcmp, st_numhash);
  /* First of all classify alphas and betas for agents */
  Classify_agents(cur_image_ptr, dag_ptr, agents_st_table_ptr);
#ifdef TRACE_CONS
  if (st_count(agents_st_table_ptr) == 0) {
    printf("------------------------------The model is propositional\n\n");
  }
#endif  
  /* Now iterate over agents */
  result = 1; test_ab = 1; test_ba = 1; 
  st_foreach_item(agents_st_table_ptr, at_st_generator, &ak_stGeneric, &av_stGeneric) {
    /* These are the current agent alphas and betas */
    alpha_lsList_ptr = Alpha_lsList_ptr(UNCAST_AGENT(av_stGeneric));
    beta_lsList_ptr = Beta_lsList_ptr(UNCAST_AGENT(av_stGeneric));
#ifdef TRACE_CONS
    if (lsLength(*alpha_lsList_ptr) == 0) {
      printf("------------------------------There is no []alpha in the assigment\n\n");
    } else if (lsLength(*beta_lsList_ptr) == 0) {
      printf("------------------------------There is no -[]beta in the assigment\n\n");
    }
#endif
    /* Iterate over betas */
    lsForEachItem(*beta_lsList_ptr, beta_lsGen, beta_lsGeneric) {
      beta_vertex_ptr = UNCAST_VERTEX_ITEM(beta_lsGeneric);

      /* Now iterate over alphas */
      lsForEachItem(*alpha_lsList_ptr, alpha_lsGen, alpha_lsGeneric) {
	alpha_vertex_ptr = UNCAST_VERTEX_ITEM(alpha_lsGeneric);

	/* Should be either ESAT(alpha_i & -beta_j) or ESAT(-alpha_i & beta_j) */
	/* Try to skip both tests */
	if (Use_memoT()) {
	  /* trying to retrieve result for ESAT(alpha_i & -beta_j) */
          Inc_param(Image_monitor_ptr(cur_image_ptr), CACHE_ACCESS);                       
	  result = Test_memoT(V_value(alpha_vertex_ptr), V_value(beta_vertex_ptr)); 
	  if (result == TEST_SAT) {
            Inc_param(Image_monitor_ptr(cur_image_ptr), CACHE_SUCCESS); 
	    result = 1; test_ab = 0; test_ba = 0;  /* ESAT(alpha_i & -beta_j) no more testing! */
	  } else {
	    if (result == TEST_UNSAT) {
              Inc_param(Image_monitor_ptr(cur_image_ptr), CACHE_SUCCESS); 
	      result = 0; test_ab = 0;             /* -ESAT(alpha_i & -beta_j) so skip this test */
	    } else {
	      result = 1; test_ab = 1;             /* do test ESAT(alpha_i & -beta_j) */
	    }
	    /* trying to retrieve result for ESAT(-alpha_i & beta_j) */
	    result = Test_memoT(V_value(beta_vertex_ptr), V_value(alpha_vertex_ptr));
            Inc_param(Image_monitor_ptr(cur_image_ptr), CACHE_ACCESS); 
	    if (result == TEST_SAT) {
              Inc_param(Image_monitor_ptr(cur_image_ptr), CACHE_SUCCESS); 
	      result = 1; test_ab = 0; test_ba = 0; /* ESAT(-alpha_i & beta_j) no more testing! */

	    } else {
	      if (result == TEST_UNSAT) {
                Inc_param(Image_monitor_ptr(cur_image_ptr), CACHE_SUCCESS);       
		result = 0; test_ba = 0;            /* -ESAT(-alpha_i & beta_j) so skip this test */
	      } else { 
		result = 1; test_ba = 1;            /* do test ESAT(-alpha_i & beta_j) */
	      }
	    }
	  }
	}

	/* Should test ESAT(alpha_i & -beta_j)? */
	if (test_ab) {
	  /* This creates to_test_lsList_ptr */
	  to_test_lsList_ptr = 
	    Merge_formulae_to_new(Make_cnf(dag_ptr, alpha_vertex_ptr, POS_CNF),
				  Make_cnf(dag_ptr, beta_vertex_ptr, NEG_CNF));
#ifdef TRACE_CONS
	  Print_simple_call(alpha_vertex_ptr, beta_vertex_ptr, to_test_lsList_ptr);
#endif
	  /* We need a new image for SATO */
	  new_image_ptr = Init_cnf_sat_rec(dag_ptr, to_test_lsList_ptr);
	  /* Passing the reference to the monitor structure */
	  Inc_param(Image_monitor_ptr(cur_image_ptr), REC_CALLS); 
	  Image_monitor_ptr(new_image_ptr) = Image_monitor_ptr(cur_image_ptr); 
	  result = Test_dp_sat_rec(new_image_ptr, dag_ptr, to_test_lsList_ptr, 
				   Test_E_consistency);
	  
	  /* Store the result for future use */
	  if (Use_memoT()) {
	    Store_memoT(V_value(alpha_vertex_ptr), V_value(beta_vertex_ptr), result);
	  }
	  
	  /* If the test was ok, no further test is required */
	  if (result) test_ba = 0;
	  
	  /* Cleaning up temporary allocations */
	  list_status = lsDestroy(*to_test_lsList_ptr, NULL);
	  FREE(to_test_lsList_ptr);
	  Clear_image(new_image_ptr);
#ifdef TRACE_CONS
	  printf("Which is %s\n\n\n", (result ? "consistent" : "unconsistent"));
#endif
	}

	/* Should test ESAT(-alpha_i & beta_j)? */
	if (test_ba) {
	  /* This creates to_test_lsList_ptr */
	  to_test_lsList_ptr = 
	    Merge_formulae_to_new(Make_cnf(dag_ptr, alpha_vertex_ptr, POS_CNF),
				  Make_cnf(dag_ptr, beta_vertex_ptr, NEG_CNF));
#ifdef TRACE_CONS
	  Print_simple_call(beta_vertex_ptr, alpha_vertex_ptr, to_test_lsList_ptr);
#endif
	  /* We need a new image for SATO */
	  new_image_ptr = Init_cnf_sat_rec(dag_ptr, to_test_lsList_ptr);
	  /* Passing the reference to the monitor structure */
	  Inc_param(Image_monitor_ptr(cur_image_ptr), REC_CALLS); 
	  Image_monitor_ptr(new_image_ptr) = Image_monitor_ptr(cur_image_ptr); 
	  result = Test_dp_sat_rec(new_image_ptr, dag_ptr, to_test_lsList_ptr, 
				   Test_E_consistency);
	  
	  /* Store the result for future use */
	  if (Use_memoT()) {
	    Store_memoT(V_value(beta_vertex_ptr), V_value(alpha_vertex_ptr), result);
	  }
	  
	  /* Cleaning up temporary allocations */
	  list_status = lsDestroy(*to_test_lsList_ptr, NULL);
	  FREE(to_test_lsList_ptr);
	  Clear_image(new_image_ptr);
#ifdef TRACE_CONS
	  printf("Which is %s\n\n\n", (result ? "consistent" : "unconsistent"));
#endif
	}
	
	/* Unconsistent? */
	if (!result) {
	  lsFinish(alpha_lsGen);
	  break; /* break the loop on alphas */
	} 
      } /* end of the loop on alphas */
      
      /* Unconsistent? */
      if (!result) {
	lsFinish(beta_lsGen);
	break; /* break the loop on betas */
      }
    } /* end of the loop on betas */
    
    /* Unconsistent? */
    if (!result) {
      st_free_gen(at_st_generator);
      break; /* break the loop on agents */
    }
  } /* end of the loop on agents */
  
  /* Cleaning up the agents table */
  st_foreach_item(agents_st_table_ptr, at_st_generator, &ak_stGeneric, &av_stGeneric) {
    Clear_agent(av_stGeneric);
  }
  st_free_table(agents_st_table_ptr);    
#ifdef TRACE_CONS
  if (result)
    printf("------------------------------The model is consistent in K\n\n");
  else
    printf("------------------------------The model is not consistent in K\n\n");
#endif
  
  return result;
}



