/*  *********************************************************************** */
/*  *********************************************************************** */

/*  FILE set.c - *SAT 1.0 */
/*  Memoizing with sets of sets */

/*  *********************************************************************** */
/*  *********************************************************************** */

#include "set.h"

/* Memoizing */
static memo_t *SET_memo_ptr;

/* *********************************************************************** */
/* Initialize and clear memoizing */
/* *********************************************************************** */
void Init_memoS(int dim)
{
  int i;
  SET_memo_ptr = ALLOC(memo_t, 1);
  /* Initialize data and dependencies sets */
  SET_memo_ptr->memo_array = ALLOC(var_set_t*, (2 * dim + 1));
  SET_memo_ptr->dep_array = ALLOC(var_set_t*, (dim + 1));
  SET_memo_ptr->dep_beta = ALLOC(int, WINDOW_SIZE);
  for (i = 0; i <= dim; i++) {
    SET_memo_ptr->memo_array[i] = var_set_new(WINDOW_SIZE);
    SET_memo_ptr->dep_array[i] = var_set_new(WINDOW_SIZE);
  }
  for (i = dim + 1; i <= (2 * dim); i++) {
    SET_memo_ptr->memo_array[i] = var_set_new(WINDOW_SIZE);
  }
  SET_memo_ptr->cur_set = -1;
  SET_memo_ptr->cur_dep = -1;
  /* Record the dimensions */
  SET_memo_ptr->dim = dim;
  /* All flags disabled */
  for (i = 0; i < MAX_FLAGS; i++)  
    SET_memo_ptr->flags[i] = 0;
  /* All stats are empty */
  for (i = 0; i < MAX_STATS; i++)  
    SET_memo_ptr->stats[i] = 0;
  /* Fast growth */
  SET_memo_ptr->growth = 0;

  /* Compute allocated memory (bytes) */
  SET_memo_ptr->stats[MEMORY] = 
    sizeof(memo_t) +                                     /* The structure */
    (3 * SET_memo_ptr->dim *                             /* Sets and dependencies */
     (sizeof(var_set_t) + WINDOW_SIZE / sizeof(int))) +
    sizeof(int) * WINDOW_SIZE;                           /* Dependency labels */ 

  return;
}

void Clear_memoS(int dim)
{
  int i;
  for (i = 0; i <= (2 * dim); i++) 
    var_set_free(SET_memo_ptr->memo_array[i]);
  FREE(SET_memo_ptr);
  return;
}

void Force_no_memoS()
{
  SET_memo_ptr = NIL(memo_t);
  return;
}


/* *********************************************************************** */
/* Accessing the structure */
/* *********************************************************************** */

int Use_memoS()
{
  return (SET_memo_ptr != NIL(memo_t));
}

void Setf_memoS(int flag)
{
  SET_memo_ptr->flags[flag] = 1;
  return;
}

void Clearf_memoS(int flag)
{
  SET_memo_ptr->flags[flag] = 0;
  return;
}

int Getf_memoS(int flag)
{
  return SET_memo_ptr->flags[flag];
}

int Stats_memoS(int stat)
{
  return SET_memo_ptr->stats[stat];
}

lsList *Test_memoS(lsList *alpha_lsList_ptr, lsList *beta_lsList_ptr)
{
  var_set_t *witness_var_set_ptr;
  lsGen     t_lsGen;
  lsGeneric vertex_lsGeneric;
  lsStatus  list_status;
  int       alpha_value, beta_value, i;

  if (SET_memo_ptr->cur_set != -1) {
    /* Create a witness, fill it with all the elements */
    witness_var_set_ptr = var_set_new(WINDOW_SIZE);
    for (i = 0; i < WINDOW_SIZE; i++) 
      var_set_set_elt(witness_var_set_ptr, i);
    
    /* Filter the witness through the alphas */
    lsForEachItem(*alpha_lsList_ptr, t_lsGen, vertex_lsGeneric) {
      alpha_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
      var_set_and(witness_var_set_ptr, 
		  witness_var_set_ptr, SET_memo_ptr->memo_array[alpha_value]);
    }
    
    /* If at least one element of the witness is set, go on 
       otherwise fail */
    if (var_set_is_empty(witness_var_set_ptr)) {
      var_set_free(witness_var_set_ptr);
      return beta_lsList_ptr;
    }
    
    /* Now try to discard as many betas as possible */
    lsForEachItem(*beta_lsList_ptr, t_lsGen, vertex_lsGeneric) {
      ++SET_memo_ptr->stats[MEMO_ACCESS];
      beta_value = 
	V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric)) + 
	SET_memo_ptr->dim;
      /* Look for a correspondence witness <--> betas */
      i = 0;
      while ((i < WINDOW_SIZE) &&
	     (!(var_set_get_elt(witness_var_set_ptr, i) &&
		var_set_get_elt(SET_memo_ptr->memo_array[beta_value], i)))) ++i;
      /* If the end was not reached then the current beta can be discarded */
      if (i != WINDOW_SIZE) {
	++SET_memo_ptr->stats[MEMO_SUCCESS]; 
	list_status = lsDelBefore(t_lsGen, &vertex_lsGeneric);
      }
    }

    /* Destroy temporary witness */
    var_set_free(witness_var_set_ptr);
  }
  
  return beta_lsList_ptr;
}

int Trash_subsets(int cur_set);

void Store_memoS(lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr)
{
  lsGen     t_lsGen;
  lsGeneric vertex_lsGeneric;
  lsStatus  list_status;
  int       alpha_value, beta_value, i, cur_set;

  if (!Getf_memoS(COMPACT_STORAGE)) {
    ++(SET_memo_ptr->stats[MEMO_STORAGE]);

    /* Increment the current set modulo WINDOW_SIZE */
    (SET_memo_ptr->cur_set)++;
    if (SET_memo_ptr->cur_set == WINDOW_SIZE) {
      SET_memo_ptr->cur_set = 0;
      /* Growth is now slow */
      SET_memo_ptr->growth = 1;
    }
    
    /* If slow growth and TRASH_SUBSETS are set then 
       try to trash subsets only */
    if ((Getf_memoS(TRASH_SUBSETS) && (SET_memo_ptr->growth))) {
      cur_set = Trash_subsets(SET_memo_ptr->cur_set);
      /* If no subset to trash was found... */
      if (cur_set == WINDOW_SIZE) 
	/* ...return to fast growth */
	SET_memo_ptr->growth = 0;
      else 
	/* ... else update the current set */
	SET_memo_ptr->cur_set = cur_set;
    }
    
    /* Be sure to clear everything before setting */
    for (i = 0; i <= (2 * SET_memo_ptr->dim); i++) 
      var_set_clear_elt(SET_memo_ptr->memo_array[i],SET_memo_ptr->cur_set);
    
    /* Set alphas properly */
    lsForEachItem(*alpha_lsList_ptr, t_lsGen, vertex_lsGeneric) {
      alpha_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
      var_set_set_elt(SET_memo_ptr->memo_array[alpha_value],
		      SET_memo_ptr->cur_set);
    }

  } 
  
  /* Set the beta properly (either for a new entry or for an old one) */
  beta_value = V_value(beta_vertex_ptr) + SET_memo_ptr->dim;
  var_set_set_elt(SET_memo_ptr->memo_array[beta_value],
		  SET_memo_ptr->cur_set);

  return;
}
  

int Trash_subsets(int cur_set)
{
  var_set_t *witness_var_set_ptr;
  int       *modal_atoms;
  int       max_atoms, i, run_set;

  modal_atoms = ALLOC(int, (2 * SET_memo_ptr->dim + 1));
  run_set = cur_set;
  while (run_set < WINDOW_SIZE) {
    max_atoms = 0;
    /* Collect all the atoms of run_set */
    for (i = 0; i <= (2 * SET_memo_ptr->dim); i++) 
      if (var_set_get_elt(SET_memo_ptr->memo_array[i], run_set)) 
	modal_atoms[max_atoms++] = i;

    /* Create a witness, fill it with all the elements but the current */
    witness_var_set_ptr = var_set_new(WINDOW_SIZE);
    for (i = 0; i < WINDOW_SIZE; i++) 
      if (i != run_set) var_set_set_elt(witness_var_set_ptr, i);

    /* Now look for possible supersets in the structure 
       (other than run_set itself!)*/
    for (i = 0; i < max_atoms; i++) 
      var_set_and(witness_var_set_ptr, 
		  witness_var_set_ptr, SET_memo_ptr->memo_array[modal_atoms[i]]);

    /* If the resulting set is not empty, supersets were found
       so run_set can be trashed */
    if (!var_set_is_empty(witness_var_set_ptr)) {
      var_set_free(witness_var_set_ptr);
      break;
    }
	
    /* No supersets found, try with the next set */
    var_set_free(witness_var_set_ptr);
    ++run_set;
  }
      
  FREE(modal_atoms);
  return run_set;
}

lsList *Test_depS(lsList *alpha_lsList_ptr)
{
  var_set_t *witness_var_set_ptr, *alphas_var_set_ptr;
  lsGen     t_lsGen;
  lsGeneric vertex_lsGeneric;
  lsStatus  list_status;
  int       alpha_value, i;

  if (SET_memo_ptr->cur_dep != -1) {
    /* Create a witness, fill it with all the elements */
    witness_var_set_ptr = var_set_new(WINDOW_SIZE);
    for (i = 0; i < WINDOW_SIZE; i++) 
      var_set_set_elt(witness_var_set_ptr, i);
    
    alphas_var_set_ptr = var_set_new(SET_memo_ptr->dim + 1);
    /* Put the alpha list into a set */
    lsForEachItem(*alpha_lsList_ptr, t_lsGen, vertex_lsGeneric) {
      alpha_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
      var_set_set_elt(alphas_var_set_ptr, alpha_value);
    }
    /* Filter the witness through the complement of the alphas */
    for (i = 0; i <= SET_memo_ptr->dim; i++) 
      if (!var_set_get_elt(alphas_var_set_ptr, i)) 
	var_set_and(witness_var_set_ptr, 
		    witness_var_set_ptr, SET_memo_ptr->dep_array[i]);

    /* If at least one element of the witness is set, go on 
       otherwise fail */
    if (var_set_is_empty(witness_var_set_ptr)) {
      var_set_free(alphas_var_set_ptr);
      var_set_free(witness_var_set_ptr);
      return alpha_lsList_ptr;
    }

#ifdef TRACE_DEP
    printf("The set "); Print_alpha(alpha_lsList_ptr); 
    printf("contains dependencies\n");
#endif

    /* Now try to discard as many alphas as possible */
    for (i = 0; i < WINDOW_SIZE; i++) 
      if (var_set_get_elt(witness_var_set_ptr, i)) 
	var_set_clear_elt(alphas_var_set_ptr, SET_memo_ptr->dep_beta[i]);

    /* Really discard alphas (from the list) */
    lsForEachItem(*alpha_lsList_ptr, t_lsGen, vertex_lsGeneric) {
      ++(SET_memo_ptr->stats[DEP_ACCESS]);
      alpha_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
      if (!var_set_get_elt(alphas_var_set_ptr, alpha_value)) {
	++(SET_memo_ptr->stats[DEP_SUCCESS]);
	list_status = lsDelBefore(t_lsGen, &vertex_lsGeneric);
      }
    }
    
    /* Destroy temporary allocations */
    var_set_free(alphas_var_set_ptr);
    var_set_free(witness_var_set_ptr);
  }
  
  return alpha_lsList_ptr;
}

void Store_depS(lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr)
{
  int       i, alpha_value;
  lsGen     t_lsGen;
  lsGeneric vertex_lsGeneric;

  ++(SET_memo_ptr->stats[DEP_STORAGE]);

  /* Increment the current set modulo WINDOW_SIZE */
  (SET_memo_ptr->cur_dep)++;
  if (SET_memo_ptr->cur_dep == WINDOW_SIZE) 
    SET_memo_ptr->cur_dep = 0;

  /* Be sure to clear everything */
  for (i = 0; i <= (SET_memo_ptr->dim); i++) 
    var_set_set_elt(SET_memo_ptr->dep_array[i],SET_memo_ptr->cur_dep);
  
#ifdef TRACE_DEP
  printf("The dependency "); Print_alpha_beta(alpha_lsList_ptr, beta_vertex_ptr);
  printf("is stored\n");
#endif

  /* Set alphas and betas properly 
     (the complement of the set (alphas + beta) is stored) */
  lsForEachItem(*alpha_lsList_ptr, t_lsGen, vertex_lsGeneric) {
    alpha_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
    var_set_clear_elt(SET_memo_ptr->dep_array[alpha_value],
		      SET_memo_ptr->cur_dep);
  }
  var_set_clear_elt(SET_memo_ptr->dep_array[V_value(beta_vertex_ptr)],
		    SET_memo_ptr->cur_dep);
  /* Label the dependency set with the beta */
  SET_memo_ptr->dep_beta[SET_memo_ptr->cur_dep] = V_value(beta_vertex_ptr);

  return;
}

  
